/* Biofeedback Software for HRV (C) TU/e and Loe Feijs 2010-2017
 * The software works with an Arduino and PPG HW on serial port (GEERTCODE.ino).
 * Warning: you may need to adjust the serial port number [0], [1], or [2].
 * Logfiles are written with RR intervals in msec.
 * Warning: not validated for medical applications.
 
 * The main classes are: 
 * RRport (encapsulates derial port handling)
 * RRparser (extracts RR = beat-to-beat interval) 
 * RRdisplay (traditional tachogram, in green)
 * HRVdisplay (similar to HRV plot, blue, upsdtairs)
 * CIRCLEdisplay (modified HRV plot, also as flowers or broken hearts).
 
 * Reference: Feijs, L.M.G., Langereis, G.R. & Boxtel, van, G.J.M. (2010). 
 * Designing for heart rate and breathing movements. 
 * Proceedings of the 6th International Workshop on DeSForM 2010 (pp. 57-67). 
 * Eindhoven: Koninklijke Philips Electronics.
 */
 
import processing.serial.*;
import java.util.*;
import java.text.*;
 
int TECHNO=0;
int NATURE=1;
int GOTHIC=2;
int STYLE=TECHNO;

PImage flower;
PImage forest;
void iniFlower(){
  if (STYLE==NATURE){
    flower = loadImage("bloemBGblue.png"); 
    forest = loadImage("fantasy-forest-wallpaper.jpg");
  }
  if (STYLE==GOTHIC){
    flower = loadImage("heartBGblueBIS.png");
    forest = loadImage("gothamcity.jpg");
  }
}
boolean blueP(color c){
  //blue test predicate
  return 10+red(c)+green(c)<=blue(c);
}
void mkFlower(PImage fl, int x, int y, int p){
  //p between 0 and 100
  //center around x,y
  //fl eg flower
  int schaal=3+(100-p)/2;
  int xoffset=(fl.width/schaal)/2;
  int yoffset=(fl.height/schaal)/2;
  for (int j=0; j<fl.height; j+=schaal)
    for (int i=0; i<fl.width; i+=schaal){
      color c=fl.get(i,j);
      color newc=c;
      if (STYLE==TECHNO) newc=color(p*red(c)/100,p*green(c)/100,p*blue(c)/100);
      if (STYLE==NATURE) newc=color(p*red(c)/100,10*sqrt(p)*green(c)/100,p*blue(c)/100);
      if (STYLE==GOTHIC) newc=color(p*red(c)/100,p*green(c)/100,p*blue(c)/100);
      fill(newc);
      noStroke();
      if (!blueP(c))
        //blue-screen filtering 
        rect(x+i/schaal-xoffset,y+j/schaal-yoffset,1,1);      
    } 
}
void rmFlower(PImage fl, PImage fr, int x, int y, int p){
  //rm is remove cf mkFlower
  //fl eg flower fr eg forest
  int schaal=3+(100-p)/2;
  int xoffset=(fl.width/schaal)/2;
  int yoffset=(fl.height/schaal)/2;
  for (int j=0; j<fl.height; j+=schaal)
    for (int i=0; i<fl.width; i+=schaal){
      color flc=fl.get(i,j);
      color frc=fr.get(x+i/schaal-xoffset,y+j/schaal-yoffset);
      //color newc=color(random(40),random(40),random(40));//10,10,10
      color newc=frc;
      fill(newc);
      noStroke();
      if (!blueP(flc))
        //blue-screen filtering 
        rect(x+i/schaal-xoffset,y+j/schaal-yoffset,1,1);      
    } 
}


class RRport { 
  private Serial myPort;
  private int[] buffer = new int[100];
  private int   bufcnt = 0;
  RRport(PApplet parent) {  
    String portName = Serial.list()[2]; 
    //com7, arduino, lower usb right (loe's laptop) 
    //of bovenste, front, loe's pc hg3.53 
    println(Serial.list());
    myPort = new Serial(parent, portName, 19200);
  } 
  void putbuf(int c) {
    if(bufcnt>=100-1) println("OVERFLOW buffer ERROR"); 
    buffer[bufcnt++]=c;
  }
   int  getbuf() {
       if (bufcnt<=0) println("UNDERFLOW buffer ERROR"); 
       int v=buffer[0];
       //works like FIFO, shift rest again
       int i=0;
       while (i<bufcnt-1){ 
             buffer[i]=buffer[i+1];
             i++;
       }     //alternative: cyclic buffer, not now
       bufcnt--;
       return v; 
  }
  void step() {
    while (myPort.available()>0) {
      int inByte = int(myPort.read()); 
      putbuf(inByte);
    }     
  }
} //   RRport
RRport myport;

class RRparser {
  int val = 0;
  int avg = 1000;
  private int tmp=0;
  private int pwr=1;
  private boolean flag;
  boolean event() {
    if (flag) {
      flag = false; 
      return true;
    }  
    else return false;
  }    
  void step(){
    int b;
    print('.');
    while (myport.bufcnt>0){
      b=myport.getbuf(); 
      //print(char(b));
      //if (b==13) println("CR");
      if (b==10) println();//println("LF");
      if ((tmp>0)&&(b==10||b==13)) {
        flag = true; 
        val=tmp/1000;//ms ipv us
        if (val>600&&val<1600) avg = 9*(avg/10) + val/10;
        if (val<0.6*avg) {
          print(" adjusting sensor "); 
          flag=false;
        }
        if (val>1.5*avg) {
          print(" ADJUSTING SENSOR "); 
          flag=false;
        }             
      }
      if ((b==10||b==13)) {
        tmp=0;
      }     
      if (b!=10&&b!=13) {
         tmp=16*tmp + hex2int(b); //msb first  
      }
    }       
  }
  private int hex2int(int c) {
    if (c=='0') return 0;  
    if (c=='1') return 1;  
    if (c=='2') return 2;  
    if (c=='3') return 3;
    if (c=='4') return 4;  
    if (c=='5') return 5;  
    if (c=='6') return 6;  
    if (c=='7') return 7;
    if (c=='8') return 8;  
    if (c=='9') return 9;  
    if (c=='A') return 10; 
    if (c=='B') return 11;
    if (c=='C') return 12; 
    if (c=='D') return 13; 
    if (c=='E') return 14; 
    if (c=='F') return 15;
    return -1;
  } 
}// RRparser
RRparser myparser;

class RRdisplay {
  int cnt;
  int SIZE;//need 150 or more for 750 wide screen, set in setup
  int xbuffer[];
  int hbuffer[];
  int pbuffer[];
  RRdisplay(){
    cnt=0;
    SIZE=width/5;
    xbuffer = new int[SIZE];
    hbuffer = new int[SIZE];
    pbuffer = new int[SIZE];
  }
  void next(int RR){
    int H=(1600-RR)/10; 
    int X=5*cnt;
    if (H<0) H=0;//height of bar
    xbuffer[cnt]=X;
    hbuffer[cnt]=H;
    pbuffer[cnt]=100;
    cnt = (cnt+1)%(SIZE);
  }
  void step() {
    doforgetfulness();
    dobars();
  }
  private void doforgetfulness() {
    int Pfactor=100-1500/(5*SIZE); 
    if (Pfactor<90) Pfactor=90;
    if (Pfactor>99) Pfactor=99;
    for (int i=0;i<SIZE;i++) {
      pbuffer[i]=Pfactor*pbuffer[i]/100;
    }     
  }    
  private void dobar(int X, int P, int H){
    int Hmax = 100;
    //P is presence 0..100% 
    int R=25; 
    int G=255; 
    int B=125;
    strokeWeight(1);
    stroke(0,0,0); 
    fill(0,0,0); 
    rect(X,height-Hmax,4,Hmax);//clear old bar 
    R=(P*R)/100; 
    G=(P*G)/100; 
    B=(P*B)/100;
    stroke(R/2,G/2,B/2); 
    fill(R,G,B); 
    rect(X,height-H,4,H);//bar is 5 wide 
  }
  private void dobars(){
    for (int i=0;i<SIZE;i++) {
      dobar(xbuffer[i],pbuffer[i],hbuffer[i]);
    }     
  }
}
RRdisplay rrdisplay;

class HRVdisplay {
  int cnt;
  int SIZE;
  int xbuffer[];
  int hbuffer[];
  int pbuffer[];
  int prevpbuffer[];
  HRVdisplay(){
    cnt=0;
    SIZE=width/5;
    xbuffer = new int[SIZE];
    hbuffer = new int[SIZE];
    pbuffer = new int[SIZE];
    prevpbuffer = new int[SIZE];
  }
  void next(int std){
    int H=(150-std)/2; 
    int X=5*cnt;
    if (H<0) H=0;//height of bar
    xbuffer[cnt]=X;
    hbuffer[cnt]=H;
    pbuffer[cnt]=100;
    prevpbuffer[cnt]=100;
    cnt = (cnt+1)%(SIZE);  
  }
  void step() {
    doforgetfulness();
    dobars();
  }
  private void doforgetfulness() {
    int Pfactor=100-1500/(5*SIZE); 
    if (Pfactor<90) Pfactor=90;
    if (Pfactor>99) Pfactor=99;
    for (int i=0;i<SIZE;i++) {
      prevpbuffer[i]=pbuffer[i];
      pbuffer[i]=Pfactor*pbuffer[i]/100;
    }     
  }    
  private void doflower(int X, int Y, int P){
    mkFlower(flower,X,Y,P);
  }
  private void eraseflower(int X, int Y, int P){
    if (P<100) rmFlower(flower,forest,X,Y,P);
  }
  private void dobar(int X, int P, int H){
    int Hmax = 100;
    //P is presence 0..100% 
    int R=25; 
    int G=175; 
    int B=255;
    strokeWeight(1);
    stroke(0,0,0); 
    fill(0,0,0); 
    rect(X,0,4,Hmax);//clear old bar 
    R=(P*R)/100; 
    G=(P*G)/100; 
    B=(P*B)/100;
    stroke(R/2,G/2,B/2); 
    fill(R,G,B); 
    rect(X,0,4,H);//bar is 5 wide 
  }
  private void dobars(){
    for (int i=0;i<SIZE;i++) {
      //eraseflower(xbuffer[i],hbuffer[i],90*prevpbuffer[i]/100);
    }    
    for (int i=0;i<SIZE;i++) {
      dobar(xbuffer[i],pbuffer[i],hbuffer[i]);
      //doflower(xbuffer[i],hbuffer[i],90*pbuffer[i]/100);
    }     
  }
}
HRVdisplay hrvdisplay;

class CIRCLEdisplay {
  int prevRR;
  int cnt;
  int SIZE;//need 150 or more for 750 wide screen, set in setup
  int xbuffer[];
  int ybuffer[];
  int pbuffer[];       
  int prevpbuffer[];
  int fbuffer[]; //foppen, IE DECREASING COUNTER, "BUDGET"TO GO
  CIRCLEdisplay(){
    prevRR=1000;
    cnt=0;
    SIZE=width/5;
    xbuffer = new int[SIZE];
    ybuffer = new int[SIZE];
    pbuffer = new int[SIZE];
    prevpbuffer = new int[SIZE];
    fbuffer = new int[SIZE];
    for (int i=0;i<SIZE;i++) {
      xbuffer[i]=width/2;
      ybuffer[i]=height/2;
      pbuffer[i]=0;
      prevpbuffer[i]=0;
      fbuffer[i]=0;
    } 
  }
  void next(int RRavg, int RR){
    int x=(RR - RRavg) / 3;
    int y=-(prevRR - RRavg) / 3;
    int xRot =  (7*x)/10 + (7*y)/10;// rotate -45 
    int yRot = -(7*x)/10 + (7*y)/10;//7/10 = sin 45
    xRot = 2*xRot;//stretch horizontally
    prevRR = RR;
    xbuffer[cnt]=width/2+xRot;
    ybuffer[cnt]=height/2+yRot;
    pbuffer[cnt]=100;
    prevpbuffer[cnt]=100;
    fbuffer[cnt]=SIZE/2;
    cnt = (cnt+1)%(SIZE);  
  }
  void step() {
    doforgetfulness();
    if (STYLE==TECHNO) doblocksandlines();
    else doflowersandlines();
  }
  private void doforgetfulness() {
    int Pfactor=90;
    for (int i=0;i<SIZE;i++) {
      if (fbuffer[i]>0){
        prevpbuffer[i]=pbuffer[i];
        pbuffer[i]=Pfactor*pbuffer[i]/100;
        fbuffer[i]--;
      }     
    }  
  }
  private void doflower(int X, int Y, int P){
    if (P>10)
      mkFlower(flower,X,Y,P);
    //else do not write very dark flower
  }
  private void eraseflower(int X, int Y, int P){
    if (P<100) rmFlower(flower,forest,X,Y,P);
  }
  private void doblock(int X, int Y, int P){
    int R=255; 
    int G=25; 
    int B=25;
    R=(P*R)/100; 
    G=(P*G)/100; 
    B=(P*B)/100;
    strokeWeight(1); 
    stroke(R/2,G/2,B/2); 
    fill(R,G,B); 
    rect(X,Y,6,6);//block is 7 wide 
  }
  private void doline(int X, int Y, int PX, int PY, int P){
    int Hmax = 160;
    //P is presence 0..100% 
    int R=255; 
    int G=25; 
    int B=25;
    R=(P*R)/100; 
    G=(P*G)/100; 
    B=(P*B)/100;
    strokeWeight(1); 
    stroke(R,G,B); 
    line(X+2,Y+2,PX+2,PY+2); 
  }
  private void dogreenline(int X, int Y, int PX, int PY, int P){
    int Hmax = 160;
    //actually it is red now
    //P is presence 0..100% 
    int R=150; 
    int G=45; 
    int B=60;
    R=(P*R)/100; 
    G=(P*G)/100; 
    B=(P*B)/100;
    //strokeWeight(1+int(P/50)); 
    strokeWeight(1);
    stroke(R,G,B); 
    if (STYLE==TECHNO) line(X+2,Y+2,PX+2,PY+2); 
    if (STYLE==GOTHIC) line(X+2,Y+2,PX+2,PY+2); 
  }
  private void erasegreenline(int X, int Y, int PX, int PY, int P){
    int Hmax = 160;
    //P is presence 0..100% 
    int R=0; 
    int G=0; 
    int B=0;
    //strokeWeight(1+int(P/50)); 
    strokeWeight(1);
    //stroke(R,G,B); line(X+2,Y+2,PX+2,PY+2); 
  }
  private int prev(int i){
    if (i>0) return i-1;
    else return SIZE-1;
  }
  private int nxt(int i){
    if (i<SIZE-1) return i+1;
    else return 0;
  }
  private void doflowersandlines(){
    //erase all old flowers
    int j=cnt; 
    for (int i=0;i<SIZE;i++) {
      j=nxt(j);//write newest last
      if (fbuffer[j]>0)
        eraseflower(xbuffer[j],ybuffer[j],prevpbuffer[j]);
    }
    //after that write new flowers
    j=cnt;
    for (int i=0;i<SIZE;i++) {
      j=nxt(j);//write newest last
      if (fbuffer[j]>0 && fbuffer[prev(j)]>0){
        erasegreenline(xbuffer[j],ybuffer[j],xbuffer[prev(j)],ybuffer[prev(j)],prevpbuffer[j]);
        dogreenline(xbuffer[j],ybuffer[j],xbuffer[prev(j)],ybuffer[prev(j)],pbuffer[j]);
      }
      if (fbuffer[j]>0){
        doflower(xbuffer[j],ybuffer[j],pbuffer[j]);
      }
    }     
  }
  private void doblocksandlines(){
    int j=cnt; 
    for (int i=0;i<SIZE;i++) {
      j=nxt(j);//write newest last
      if (fbuffer[j]>0)
        doblock(xbuffer[j],ybuffer[j],pbuffer[j]);
      if (fbuffer[j]>0 && fbuffer[prev(j)]>0)
        doline(xbuffer[j],ybuffer[j],xbuffer[prev(j)],ybuffer[prev(j)],pbuffer[j]);
    }     
  }
}
CIRCLEdisplay circledisplay;

boolean settedup=false;
PrintWriter output;
PFont myFont; 
void setup()
{ 
  int bars=150;//was 75
  iniFlower();
  frameRate(10);
  myFont = createFont("TimesNewRomanPSMT",16);
  textFont(myFont);
  //size(7*bars,5*bars);//processing 2
  size(1050,750);//processing 3 @#$%
  myport = new RRport(this);
  myparser = new RRparser();
  rrdisplay = new RRdisplay();
  hrvdisplay = new HRVdisplay();
  circledisplay = new CIRCLEdisplay();

  SimpleDateFormat sdf = new SimpleDateFormat("-yyyy-MM-dd/HH-mm"); 
  Date now = new Date(); 
  println("log to logfiles"+sdf.format(now));
  output = createWriter("logfiles"+sdf.format(now)+".txt");

  settedup=true;
  fill(10,10,10);
  rect(0,0,width,height);
  if (STYLE==NATURE) background(forest);
  if (STYLE==GOTHIC) background(forest);
}

int RRavg = 1200;
int RRstd = 50;
int count = 0;
int time = 0;
void draw() {
  count++;
  count=count%10;
  if (count==9) output.flush();
  time++;
}

void serialEvent(Serial p) { if (settedup){
  //print("!");
  //zat allemaal eerst in draw
  myport.step();
  myparser.step();
  if (time/10 > 12000) stop();//was 120
  if (myparser.event()) {
    int RR = myparser.val; 
    print(" RR="); 
    print(RR); 
    output.println(RR);
    RRavg = int((23 * RRavg + 1*RR)/24); 
    print(" AVG="); 
    print(RRavg);
    RRstd = int((15.0 * RRstd + abs(RR - RRavg)) / 16.0); 
    print(" STD="); 
    print(RRstd);//beter met kwadraat

    if (STYLE==TECHNO) rrdisplay.next(RR);
    if (STYLE==TECHNO) hrvdisplay.next(RRstd);
    circledisplay.next(RRavg,RR);
    rrdisplay.step();
    hrvdisplay.step();
    circledisplay.step();
    float BTopt = 10.00; //assumed resonant breathing rate in seconds
    int advice = int(BTopt*1000/RRavg); //idem in heart beats
    fill(30); 
    stroke(0); 
    rect(0,height-16,16,2*16);
    fill(150); 
    text(Integer.toString(advice),0, height);
  }
}}

public void stop() {
  output.flush(); // Writes the remaining data to the file
  output.close(); // Finishes the file
  println("THANK YOU, GOODBYE");
  exit(); // Stops the program 
}

void keyPressed() {

}